/**
 * @file
 * @author  Hannes Jung <hannes.jung@desy.de>
 *
 * @section LICENSE
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details at
 * https://www.gnu.org/copyleft/gpl.html
 *
 *
 */
 
 
#ifndef TMDlib_H
#define TMDlib_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <cstring>
#include <math.h>
#include <vector>
#include <iostream>
#include <map>

using namespace std;

namespace TMDlib {

  double ipow(double,int);
  extern int iipow(int,int);


class TMD {
   public:
      TMD();  // This is the constructor
      
      void TMDinit(const string name, int irep, int imode  ) ;
      void TMDinit(const string name, int irep ) ;
      void TMDinit(const string name) ;
      void TMDinit(int & iset) ;
      
      /// Strip leading and trailing spaces (not in-place) from LHAPDF
      inline std::string trim(const std::string& s) {
        const size_t firstnonspacepos = s.find_first_not_of(" ");
        const size_t lastnonspacepos = s.find_last_not_of(" ");
        if (firstnonspacepos == std::string::npos) return "";
        return s.substr(firstnonspacepos, lastnonspacepos-firstnonspacepos+1);
      }

  
      // TMDpdf 1st to return TMD value at x,xbar,kt,mu as a vector
      vector<double> TMDpdf( double x, double xbar, double kt, double mu) ;
      // TMDpdf 3rd to return TMD value at x,xbar,kt,mu as a vector
      void TMDpdf( double x, double xbar, double kt, double mu, vector<double>& xpq) ;
      // TMDpdf 5th to return TMD value at x,xbar,kt,mu for flavor
      void TMDpdf( double x, double xbar, double kt, double mu, double& uval, double& dval, double& sea, double& charm, double& bottom, double& gluon );
      // TMDpdf 7th to return TMD value at x,xbar,kt,mu for flavor
      void TMDpdf( double x, double xbar, double kt, double mu, double& up, double& ubar, double& dn, double& dbar, double& strange,  double& sbar, double& charm, double& cbar, double& bottom, double& bbar, double& gluon );
      // main TMD routine
      void TMDdensity( double x, double xbar, double kt, double mu, double& up, double& ubar, double& dn, double& dbar, double& strange,  double& sbar, double& charm, double& cbar, double& bottom, double& bbar, double& gluon );
  
      double TMDalphas(double mu);
  
      double TMDgetLam4( );
      int TMDgetNf( );
      int TMDgetNumMembers( );
      int TMDgetOrderAlphaS( );
      int TMDgetOrderPDF( );
      double TMDgetXmin();
      double TMDgetXmax();
      double TMDgetQ2min();
      double TMDgetQ2max();
      double TMDgetQmin();
      double TMDgetQmax();
      double TMDgetKtmin();
      double TMDgetKtmax();
      string TMDgetDesc();
      string TMDgetScheme();
      int TMDnumberPDF(string name);
      string TMDstringPDF(int index);    
      void TMDinfo(const string name);
      string TMDgetExtrapolation_Q2();
      string TMDgetExtrapolation_x();
      string TMDgetExtrapolation_kt();
      
      bool DoesFileExist (const string name);
      
      double get_key_val_as_double(const string name);
      int get_key_val_as_int(const string name);


      //double Cdhfint(int narg, double* arg[500], int* nent[500], double* ent[500], double* table[1000000]);  
      //double Cdhfint(int narg, double arg[500], int nent[500], double ent[500], double table[1000000]);  
      double Cdhfint(int narg, double arg[], int nent[], double ent[], double table[]);  
      void polint(double xa[], double ya[], int n, double x, double& y, double& dy) ;
      void polin3(double x1a[], double x2a[], double x3a[], double ya[][1][4], int m, int n, int o, double x1, double x2, double x3, double& y, double& dy); 
// PB stuff
      void allFlavuPDF(double x, double kt, double p, double& up, double& ubar, double& dn, double& dbar, double& strange, double& sbar, double& charm, double& cbar, double& bottom, double& bbar, double& top, double& tbar, double& phot, double& gluon ) ;
      void allFlavuPDF_old(double x, double kt, double p, double& up, double& ubar, double& dn, double& dbar, double& strange, double& sbar, double& charm, double& cbar, double& bottom, double& bbar, double& top, double& tbar, double& phot, double& gluon ) ;
      void allFlavuPDFn(double x, double kt, double p, double& up, double& ubar, double& dn, double& dbar, double& strange, double& sbar, double& charm, double& cbar, double& bottom, double& bbar, double& top, double& tbar, double& phot, double& gluon ) ;
      void PaviaTemp(double x, double kt, double p, double& up, double& ubar, double& dn, double& dbar, double& strange, double& sbar, double& charm, double& cbar, double& bottom, double& bbar, double& top, double& tbar, double& phot, double& gluon ) ;
                  
// GBW 
      void gbwuPDFlight(double x, double kt, double& gluon ) ;
      void gbwuPDFcharm(double x, double kt, double& gluon ) ;
  // double blueml(double x, double kt, double p ) ;


// SBRS
      void sbrsPDF(int irep, int imode, double x, double kt, double p, double& uval, double& dval, double& sea, double& charm);

// Kutak Sapeta
      void ksPDF(double x, double kt, double p, double& upl, double& dn, double& sea, double& charm, double& bottom, double& gluon);
      void ksuPDFgrid(int kf, double x, double kt , double p, double& up, double& dn, double& sea, double& charm, double& bottom, double& gluon ) ;
      void ksDLCPDF(double x, double kt, double p, double& up, double& ubar, double& down, double& dbar, double& strange, double& sbar, double& charm,double& cbar, double& bottom, double& bbar, double& gluon);
      void ksBHKSPDF(double x, double kt, double p, double& up, double& ubar, double& down, double& dbar, double& strange, double& sbar, double& charm, double& cbar, double& bottom, double& bbar, double& gluon);

// M. Echevaria TMDgluon  
      void tmd_ME(double x, double kt , double p, double& uval, double& dval, double& sea, double& charm, double& bottom, double& gluon ) ;
      
   private:
      int iset, irep, imode ;
      // const string pdfpath;
      bool first ;
      int ncallTMDdensity;
    
      int NewFormat;
      string TMD_Name, TMD_Dir, TMD_Mem;

      double qscal, qg0, q2;
      int n1min, n1max, n2min, n2max, n3min, n3max, ncall;
      int iqqbar, iglu, ikincut, ipgg, ns_sel;
      double* a = new double[153];
      double* xa= new double[3];
      double* px = new double[52];
      double* xx = new double[52];
      double* q2x = new double[52];
      double* f_grid0 = new double[132651];
      double* f_grid1 = new double[132651];
      double* f_grid2 = new double[132651];
      double* f_grid3 = new double[132651];
      double* f_grid4 = new double[132651];
      double* f_grid5 = new double[132651];
      double* f_grid6 = new double[132651];
      double* f_grid7 = new double[132651];
      double* f_grid1m = new double[132651];
      double* f_grid2m = new double[132651];
      double* f_grid3m = new double[132651];
      double* f_grid4m = new double[132651];
      double* f_grid5m = new double[132651];
      double* f_grid6m = new double[132651];
    
      std::map<string,string> TMDdict;
      std::map<int, string> TMDindex;
  

};


      extern "C" { 
     // Fortran routines
     // remember FORTRAN passes pointer (instead of values) to subroutines
     #define initalphas initalphas_
     void initalphas(int* iord, double* fr2, double* mur , double* asmur, double* mc, double* mb, double* mt); 	
     #define unpol unpol_
     void unpol(int* in , double* scale, double* scale2, double* bc, double* lam1, double* lam2, double* x_ME, double* kt_ME,double* Qf, double* result ); 
     #define blueml blueml_
     double blueml(double* x, double* kt, double* p ) ;  
    }

  

}

#endif
